# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helpers used in the PyACP examples.

These utilities can download the input files used in the PyACP examples.
"""
import dataclasses
from enum import Enum, auto
import pathlib
import shutil
import sys
import tempfile
import urllib.parse
import urllib.request

__all__ = [
    "ExampleKeys",
    "get_example_file",
    "FLAT_PLATE_SHELL_CAMERA",
    "FLAT_PLATE_SOLID_CAMERA",
    "RACE_CARE_NOSE_CAMERA_METER",
    "set_plot_theme",
]

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ansys.acp.core import Model

_EXAMPLE_REPO = "https://github.com/ansys/example-data/raw/master/pyacp/"


# _EXAMPLE_REPO = "D:\\ANSYSDev\\pyansys-example-data\\pyacp\\"


# Order of inputs: position, rotation point, orientation
FLAT_PLATE_SHELL_CAMERA = [
    (-0.0053, 0.0168, 0.0220),
    (0.0022, 0.0041, 0.0104),
    (0.3510, 0.7368, -0.5779),
]
FLAT_PLATE_SOLID_CAMERA = [
    (0.0251, 0.0144, 0.0256),
    (0.0086, 0.0041, 0.0089),
    (-0.2895, 0.9160, -0.2776),
]

# Order of inputs: position, rotation point, orientation
RACE_CARE_NOSE_CAMERA_METER = [
    (1.614, 1.154, 2.243),
    (0.450, 0.238, -0.181),
    (-0.1094, 0.9460, -0.3050),
]


@dataclasses.dataclass
class _ExampleLocation:
    directory: str
    filename: str


class ExampleKeys(Enum):
    """Keys for the example files."""

    BASIC_FLAT_PLATE_DAT = auto()
    BASIC_FLAT_PLATE_REFINED_DAT = auto()
    BASIC_FLAT_PLATE_ACPH5 = auto()
    BASIC_FLAT_PLATE_SOLID_MESH_CDB = auto()
    RACE_CAR_NOSE_ACPH5 = auto()
    RACE_CAR_NOSE_STEP = auto()
    CUT_OFF_GEOMETRY = auto()
    RULE_GEOMETRY_TRIANGLE = auto()
    THICKNESS_GEOMETRY = auto()
    MINIMAL_FLAT_PLATE = auto()
    OPTIMIZATION_EXAMPLE_DAT = auto()
    CLASS40_AGDB = auto()
    CLASS40_CDB = auto()
    MATERIALS_XML = auto()
    IMPORTED_SOLID_MODEL_ACPH5 = auto()
    IMPORTED_SOLID_MODEL_SOLID_MESH = auto()
    SNAP_TO_GEOMETRY = auto()
    CUT_OFF_GEOMETRY_SOLID_MODEL = auto()
    MINIMAL_PLATE_ACPH5 = auto()
    MINIMAL_PLATE_CDB = auto()


EXAMPLE_FILES: dict[ExampleKeys, _ExampleLocation] = {
    ExampleKeys.BASIC_FLAT_PLATE_DAT: _ExampleLocation(
        directory="basic_flat_plate_example", filename="flat_plate_input.dat"
    ),
    ExampleKeys.BASIC_FLAT_PLATE_REFINED_DAT: _ExampleLocation(
        directory="basic_flat_plate_example", filename="flat_plate_input_refined.dat"
    ),
    ExampleKeys.BASIC_FLAT_PLATE_ACPH5: _ExampleLocation(
        directory="basic_flat_plate_example", filename="flat_plate.acph5"
    ),
    ExampleKeys.BASIC_FLAT_PLATE_SOLID_MESH_CDB: _ExampleLocation(
        directory="basic_flat_plate_example", filename="solid_mesh.cdb"
    ),
    ExampleKeys.RACE_CAR_NOSE_ACPH5: _ExampleLocation(
        directory="race_car_nose", filename="race_car_nose.acph5"
    ),
    ExampleKeys.RACE_CAR_NOSE_STEP: _ExampleLocation(
        directory="race_car_nose", filename="race_car_nose.stp"
    ),
    ExampleKeys.CUT_OFF_GEOMETRY: _ExampleLocation(
        directory="geometries", filename="cut_off_geometry.stp"
    ),
    ExampleKeys.RULE_GEOMETRY_TRIANGLE: _ExampleLocation(
        directory="geometries", filename="rule_geometry_triangle.stp"
    ),
    ExampleKeys.THICKNESS_GEOMETRY: _ExampleLocation(
        directory="geometries", filename="thickness_geometry.stp"
    ),
    ExampleKeys.MINIMAL_FLAT_PLATE: _ExampleLocation(
        directory="basic_flat_plate_example", filename="minimal_model_single_ply.acph5"
    ),
    ExampleKeys.OPTIMIZATION_EXAMPLE_DAT: _ExampleLocation(
        directory="optimization_example", filename="optimization_model.dat"
    ),
    ExampleKeys.CLASS40_AGDB: _ExampleLocation(directory="class40", filename="class40.agdb"),
    ExampleKeys.CLASS40_CDB: _ExampleLocation(directory="class40", filename="class40.cdb"),
    ExampleKeys.MATERIALS_XML: _ExampleLocation(directory="materials", filename="materials.engd"),
    ExampleKeys.IMPORTED_SOLID_MODEL_ACPH5: _ExampleLocation(
        directory="imported_solid_model", filename="t-joint-ACP-Pre.acph5"
    ),
    ExampleKeys.IMPORTED_SOLID_MODEL_SOLID_MESH: _ExampleLocation(
        directory="imported_solid_model", filename="t-joint.solid.h5"
    ),
    ExampleKeys.SNAP_TO_GEOMETRY: _ExampleLocation(
        directory="geometries", filename="snap_to_geometry.stp"
    ),
    ExampleKeys.CUT_OFF_GEOMETRY_SOLID_MODEL: _ExampleLocation(
        directory="geometries", filename="cut_off_geometry_solid_model.stp"
    ),
    ExampleKeys.MINIMAL_PLATE_ACPH5: _ExampleLocation(
        directory="getting_started", filename="minimal_plate_no_matml.acph5"
    ),
    ExampleKeys.MINIMAL_PLATE_CDB: _ExampleLocation(
        directory="getting_started", filename="minimal_plate.cdb"
    ),
}


def get_example_file(example_key: ExampleKeys, working_directory: pathlib.Path) -> pathlib.Path:
    """Download an example file from the example-data repo to the working directory.

    Parameters
    ----------
    example_key
        Key for the example file.
    working_directory
        Working directory to download the example file to.
    """
    example_location = EXAMPLE_FILES[example_key]
    _download_file(example_location, working_directory / example_location.filename)
    return working_directory / example_location.filename


def _is_url(path_url: str) -> bool:  # pragma: no cover
    return urllib.parse.urlparse(path_url).scheme in ["http", "https"]


def _get_file_url(example_location: _ExampleLocation) -> str:  # pragma: no cover
    if sys.platform == "win32" and not _is_url(_EXAMPLE_REPO):
        return _EXAMPLE_REPO + "\\".join([example_location.directory, example_location.filename])
    else:
        return _EXAMPLE_REPO + "/".join([example_location.directory, example_location.filename])


def _download_file(
    example_location: _ExampleLocation, local_path: pathlib.Path
) -> None:  # pragma: no cover
    file_url = _get_file_url(example_location)
    # The URL is hard-coded to start with the example repository URL, so it is safe to use
    if _is_url(file_url):
        urllib.request.urlretrieve(file_url, local_path)  # nosec: B310
    else:
        shutil.copyfile(file_url, local_path)


def _run_analysis(model: "Model") -> None:
    """Run the model with mapdl and do a post-processing analysis.

    Uses a max strain criteria, which means strain limits have to be defined.
    This function can be called at the end of examples to verify the prepared model
    actually solves and can be postprocessed.
    """
    from ansys.mapdl.core import launch_mapdl

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir_path = pathlib.Path(tmp_dir)
        model.update()
        cdb_out_path = tmp_dir_path / "model.cdb"
        model.export_analysis_model(cdb_out_path)

        # Launch the MAPDL instance
        mapdl = launch_mapdl()
        mapdl.clear()

        # Load the CDB file into PyMAPDL
        mapdl.input(str(cdb_out_path))

        # Solve the model
        mapdl.allsel()
        mapdl.slashsolu()
        mapdl.solve()

        # Download the rst file for composite specific post-processing
        rstfile_name = f"{mapdl.jobname}.rst"
        rst_file_local_path = tmp_dir_path / rstfile_name
        mapdl.download(rstfile_name, str(tmp_dir_path))

        from ansys.acp.core.dpf_integration_helpers import get_dpf_unit_system
        from ansys.dpf.composites.composite_model import CompositeModel
        from ansys.dpf.composites.constants import FailureOutput
        from ansys.dpf.composites.data_sources import (
            CompositeDefinitionFiles,
            ContinuousFiberCompositesFiles,
        )
        from ansys.dpf.composites.failure_criteria import (
            CombinedFailureCriterion,
            MaxStrainCriterion,
        )
        from ansys.dpf.composites.server_helpers import connect_to_or_start_server

        dpf_server = connect_to_or_start_server()

        max_strain = MaxStrainCriterion()

        cfc = CombinedFailureCriterion(
            name="Combined Failure Criterion",
            failure_criteria=[max_strain],
        )

        materials_file_path = tmp_dir_path / "materials.xml"
        model.export_materials(materials_file_path)
        composite_definitions_file_path = tmp_dir_path / "ACPCompositeDefinitions.h5"
        model.export_shell_composite_definitions(composite_definitions_file_path)
        composite_model = CompositeModel(
            composite_files=ContinuousFiberCompositesFiles(
                rst=rst_file_local_path,
                composite={"shell": CompositeDefinitionFiles(composite_definitions_file_path)},
                engineering_data=materials_file_path,
            ),
            default_unit_system=get_dpf_unit_system(model.unit_system),
            server=dpf_server,
        )

        output_all_elements = composite_model.evaluate_failure_criteria(cfc)
        irf_field = output_all_elements.get_field({"failure_label": FailureOutput.FAILURE_VALUE})

        # %%
        # Release composite model to close open streams to result file.
        composite_model = None  # type: ignore

        # Close MAPDL instance
        mapdl.exit()


def set_plot_theme() -> None:
    """Set the theme options used in the examples.

    This function sets the default plotting theme options to the ones used
    in the examples. You can call this function to ensure the plots are using
    the same theme (color maps etc.) as the rendered online documentation.
    """
    # PyMAPDL and PyDPF override the default PyVista theme, so we need to import it
    # here to have a chance of re-setting it.
    import ansys.mapdl.core  # isort:skip
    import ansys.dpf.core  # isort:skip # noqa: F401
    import pyvista

    # Use the PyVista 'document' theme (white background, black text, etc.),
    # except with the 'reversed' viridis color map. This is more suited to
    # e.g. thickness plots, because the thicker parts of the model will be
    # in dark colors, instead of light colors.
    pyvista.global_theme = pyvista.themes.DocumentTheme()
    pyvista.global_theme.cmap = "viridis_r"



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses

from ansys.api.acp.v0 import tube_selection_rule_pb2, tube_selection_rule_pb2_grpc

from .._utils.array_conversions import to_1D_double_array, to_tuple_from_1D_array
from .._utils.property_protocols import ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .edge_set import EdgeSet
from .enums import status_type_from_pb
from .object_registry import register

# Workaround: these imports are needed to make sphinx_autodoc_typehints understand
# the inherited members of the Elemental- and NodalData classes.
import numpy as np  # noqa: F401 isort:skip
from ._elemental_or_nodal_data import ScalarData  # noqa: F401 isort:skip

__all__ = [
    "TubeSelectionRule",
    "TubeSelectionRuleElementalData",
    "TubeSelectionRuleNodalData",
]


@dataclasses.dataclass
class TubeSelectionRuleElementalData(ElementalData):
    """Represents elemental data for a Tube Selection Rule."""

    normal: VectorData | None = None


@dataclasses.dataclass
class TubeSelectionRuleNodalData(NodalData):
    """Represents nodal data for a Tube Selection Rule."""


@mark_grpc_properties
@register
class TubeSelectionRule(CreatableTreeObject, IdTreeObject):
    """Instantiate a Tube Selection Rule.

    Parameters
    ----------
    name :
        Name of the Tube Selection Rule.
    edge_set :
        Edge Set defining the path of the tube.
    outer_radius :
        Outer radius of the tube.
    inner_radius :
        Inner radius of the tube.
    include_rule :
        Include or exclude area in rule. Setting this to ``False``
        inverts the selection.
    extend_endings :
        Enable this option to extend the endings of the tube.
    symmetrical_extension :
        Whether the extensions have equal length on both sides or not.
    head :
        Defines the head of the tube. Select a point nearby one of the free ends
        of the tube. Only needed if the extensions are asymmetric.
    head_extension :
        Defines the length of the extension at the head of the tube. In case of
        symmetry, this value is used for both sides.
    tail_extension :
        Defines the length of the extension at the tail of the tube. Only needed
        if the extensions are asymmetric.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "tube_selection_rules"
    _OBJECT_INFO_TYPE = tube_selection_rule_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = tube_selection_rule_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "TubeSelectionrule",
        edge_set: EdgeSet | None = None,
        outer_radius: float = 1.0,
        inner_radius: float = 0.0,
        include_rule: bool = True,
        extend_endings: bool = False,
        symmetrical_extension: bool = True,
        head: tuple[float, float, float] = (0.0, 0.0, 0.0),
        head_extension: float = 0.0,
        tail_extension: float = 0.0,
    ):
        super().__init__(name=name)
        self.edge_set = edge_set
        self.outer_radius = outer_radius
        self.inner_radius = inner_radius
        self.include_rule = include_rule
        self.extend_endings = extend_endings
        self.symmetrical_extension = symmetrical_extension
        self.head = head
        self.head_extension = head_extension
        self.tail_extension = tail_extension

    def _create_stub(self) -> tube_selection_rule_pb2_grpc.ObjectServiceStub:
        return tube_selection_rule_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    edge_set = grpc_link_property("properties.edge_set", allowed_types=EdgeSet)
    outer_radius: ReadWriteProperty[float, float] = grpc_data_property("properties.outer_radius")
    inner_radius: ReadWriteProperty[float, float] = grpc_data_property("properties.inner_radius")
    include_rule: ReadWriteProperty[bool, bool] = grpc_data_property("properties.include_rule_type")
    extend_endings: ReadWriteProperty[bool, bool] = grpc_data_property("properties.extend_endings")
    symmetrical_extension: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.symmetrical_extension"
    )
    head = grpc_data_property(
        "properties.head", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )
    head_extension: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.head_extension"
    )
    tail_extension: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.tail_extension"
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property
    # selection rules don't have solid mesh data
    elemental_data = elemental_data_property(TubeSelectionRuleElementalData)
    nodal_data = nodal_data_property(TubeSelectionRuleNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Collection, Iterable
import dataclasses

from ansys.api.acp.v0 import element_set_pb2, element_set_pb2_grpc

from .._utils.array_conversions import to_1D_int_array, to_tuple_from_1D_array
from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .enums import status_type_from_pb
from .object_registry import register

# Workaround: these imports are needed to make sphinx_autodoc_typehints understand
# the inherited members of the Elemental- and NodalData classes.
import numpy as np  # noqa: F401 isort:skip
from ._elemental_or_nodal_data import ScalarData  # noqa: F401 isort:skip

__all__ = [
    "ElementSet",
    "ElementSetElementalData",
    "ElementSetNodalData",
]


@dataclasses.dataclass
class ElementSetElementalData(ElementalData):
    """Represents elemental data for an Element Set."""

    normal: VectorData | None = None


@dataclasses.dataclass
class ElementSetNodalData(NodalData):
    """Represents nodal data for an Element Set."""


@mark_grpc_properties
@register
class ElementSet(CreatableTreeObject, IdTreeObject):
    """Instantiate an Element Set.

    Parameters
    ----------
    name :
        The name of the Element Set.
    middle_offset :
        If True, enforce that the laminate mid-plane is moved onto the reference surface.
    element_labels :
        Label of elements to be assigned to Element Set
    """

    __slots__: Iterable[str] = tuple()
    _COLLECTION_LABEL = "element_sets"
    _OBJECT_INFO_TYPE = element_set_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = element_set_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "ElementSet",
        middle_offset: bool = False,
        element_labels: Collection[int] = tuple(),
    ):
        super().__init__(name=name)
        self.middle_offset = middle_offset
        self.element_labels = element_labels

    def _create_stub(self) -> element_set_pb2_grpc.ObjectServiceStub:
        return element_set_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    middle_offset: ReadWriteProperty[bool, bool] = grpc_data_property("properties.middle_offset")
    element_labels: ReadWriteProperty[tuple[int, ...], Collection[int]] = grpc_data_property(
        "properties.element_labels",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_int_array,
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property
    elemental_data = elemental_data_property(ElementSetElementalData)
    nodal_data = nodal_data_property(ElementSetNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing_extensions import Self

from ansys.api.acp.v0 import material_pb2

from ..._grpc_helpers.property_helper import mark_grpc_properties
from .base import (
    _ISOTROPIC_PROPERTY_UNAVAILABLE_MSG,
    _ORTHOTROPIC_PROPERTY_UNAVAILABLE_MSG,
    _ConstantPropertySet,
    _PolymorphicMixin,
    _PolymorphicPropertyKwargs,
    _VariablePropertySet,
)
from .property_helper import (
    constant_material_grpc_data_property,
    variable_material_grpc_data_property,
)

__all__ = [
    "ConstantStressLimits",
    "VariableStressLimits",
]


class _StressLimitsMixin(_PolymorphicMixin):
    _DEFAULT_PB_PROPERTYSET_TYPE = material_pb2.OrthotropicStressLimitsPropertySet
    _FIELD_NAME_SUFFIX_DEFAULT = "_orthotropic"
    _FIELD_NAME_SUFFIX_BY_PB_DATATYPE = {
        material_pb2.IsotropicStressLimitsPropertySet: "_isotropic",
        material_pb2.OrthotropicStressLimitsPropertySet: "_orthotropic",
    }


_ISOTROPIC_KWARGS: _PolymorphicPropertyKwargs = {
    "available_on_pb_type": material_pb2.IsotropicStressLimitsPropertySet,
    "unavailable_msg": _ISOTROPIC_PROPERTY_UNAVAILABLE_MSG,
}
_ORTHOTROPIC_KWARGS: _PolymorphicPropertyKwargs = {
    "available_on_pb_type": material_pb2.OrthotropicStressLimitsPropertySet,
    "unavailable_msg": _ORTHOTROPIC_PROPERTY_UNAVAILABLE_MSG,
}


@mark_grpc_properties
class ConstantStressLimits(_StressLimitsMixin, _ConstantPropertySet):
    """Constant stress limits material property set."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    @classmethod
    def from_isotropic_constants(
        cls,
        *,
        effective_stress: float = 0.0,
    ) -> Self:
        """Create an isotropic stress limits property set.

        Parameters
        ----------
        effective_stress:
            Effective stress limit.

        Returns
        -------
        :
            An isotropic stress limits property set.
        """
        obj = cls(
            _pb_object=cls._create_pb_object_from_propertyset_type(
                material_pb2.IsotropicStressLimitsPropertySet
            )
        )
        obj.effective_stress = effective_stress
        return obj

    @classmethod
    def from_orthotropic_constants(
        cls,
        *,
        Xc: float = 0.0,
        Yc: float = 0.0,
        Zc: float = 0.0,
        Xt: float = 0.0,
        Yt: float = 0.0,
        Zt: float = 0.0,
        Sxy: float = 0.0,
        Syz: float = 0.0,
        Sxz: float = 0.0,
    ) -> Self:
        """Create an orthotropic stress limits property set.

        Parameters
        ----------
        Xc:
            Compressive stress limit in material 1 direction.
        Yc:
            Compressive stress limit in material 2 direction.
        Zc:
            Compressive stress limit in out-of-plane direction.
        Xt:
            Tensile stress limit in material 1 direction.
        Yt:
            Tensile stress limit in material 2 direction.
        Zt:
            Tensile stress limit in out-of-plane direction.
        Sxy:
            Shear stress limit in-plane (:math:`e_{12}`).
        Syz:
            Shear stress limit out-of-plane (:math:`e_{23}`).
        Sxz:
            Shear stress limit out-of-plane (:math:`e_{13}`).

        Returns
        -------
        :
            An orthotropic stress limits property set.
        """
        obj = cls(
            _pb_object=cls._create_pb_object_from_propertyset_type(
                material_pb2.OrthotropicStressLimitsPropertySet
            )
        )
        obj.Xc = Xc
        obj.Yc = Yc
        obj.Zc = Zc
        obj.Xt = Xt
        obj.Yt = Yt
        obj.Zt = Zt
        obj.Sxy = Sxy
        obj.Syz = Syz
        obj.Sxz = Sxz
        return obj

    effective_stress = constant_material_grpc_data_property("effective_stress", **_ISOTROPIC_KWARGS)
    Xc = constant_material_grpc_data_property("Xc", **_ORTHOTROPIC_KWARGS)
    Yc = constant_material_grpc_data_property("Yc", **_ORTHOTROPIC_KWARGS)
    Zc = constant_material_grpc_data_property("Zc", **_ORTHOTROPIC_KWARGS)
    Xt = constant_material_grpc_data_property("Xt", **_ORTHOTROPIC_KWARGS)
    Yt = constant_material_grpc_data_property("Yt", **_ORTHOTROPIC_KWARGS)
    Zt = constant_material_grpc_data_property("Zt", **_ORTHOTROPIC_KWARGS)
    Sxy = constant_material_grpc_data_property("Sxy", **_ORTHOTROPIC_KWARGS)
    Syz = constant_material_grpc_data_property("Syz", **_ORTHOTROPIC_KWARGS)
    Sxz = constant_material_grpc_data_property("Sxz", **_ORTHOTROPIC_KWARGS)


@mark_grpc_properties
class VariableStressLimits(_StressLimitsMixin, _VariablePropertySet):
    """Variable stress limits material property set."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    effective_stress = variable_material_grpc_data_property("effective_stress", **_ISOTROPIC_KWARGS)
    Xc = variable_material_grpc_data_property("Xc", **_ORTHOTROPIC_KWARGS)
    Yc = variable_material_grpc_data_property("Yc", **_ORTHOTROPIC_KWARGS)
    Zc = variable_material_grpc_data_property("Zc", **_ORTHOTROPIC_KWARGS)
    Xt = variable_material_grpc_data_property("Xt", **_ORTHOTROPIC_KWARGS)
    Yt = variable_material_grpc_data_property("Yt", **_ORTHOTROPIC_KWARGS)
    Zt = variable_material_grpc_data_property("Zt", **_ORTHOTROPIC_KWARGS)
    Sxy = variable_material_grpc_data_property("Sxy", **_ORTHOTROPIC_KWARGS)
    Syz = variable_material_grpc_data_property("Syz", **_ORTHOTROPIC_KWARGS)
    Sxz = variable_material_grpc_data_property("Sxz", **_ORTHOTROPIC_KWARGS)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from ansys.api.acp.v0 import material_pb2

from ..._grpc_helpers.property_helper import mark_grpc_properties
from ...base import TreeObject
from .base import _ConstantPropertySet, _VariablePropertySet
from .property_helper import (
    constant_material_grpc_data_property,
    variable_material_grpc_data_property,
)

__all__ = ["ConstantFabricFiberAngle", "VariableFabricFiberAngle"]


class _FabricFiberAngleMixin:
    _DEFAULT_PB_PROPERTYSET_TYPE = material_pb2.FabricFiberAnglePropertySet


@mark_grpc_properties
class ConstantFabricFiberAngle(_FabricFiberAngleMixin, _ConstantPropertySet):
    """Constant fabric fiber angle material property set.

    Defines the rotation angle between the material coordinate system and the fiber direction.
    Only used for shear dependent material properties which are provided by Material Designer.
    """

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        fabric_fiber_angle: float = 0.0,
        _parent_object: TreeObject | None = None,
        _attribute_path: str | None = None,
    ):
        super().__init__(_parent_object=_parent_object, _attribute_path=_attribute_path)
        if _parent_object is not None:
            return
        self.fabric_fiber_angle = fabric_fiber_angle

    fabric_fiber_angle = constant_material_grpc_data_property("fabric_fiber_angle")


@mark_grpc_properties
class VariableFabricFiberAngle(_FabricFiberAngleMixin, _VariablePropertySet):
    """Variable fabric fiber angle material property set.

    Defines the rotation angle between the material coordinate system and the fiber direction.
    Only used for shear dependent material properties which are provided by Material Designer.
    """

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    fabric_fiber_angle = variable_material_grpc_data_property("fabric_fiber_angle")



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Base classes for tree objects backed via gRPC API."""
from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Iterable, Iterator
import contextlib
from dataclasses import dataclass
import typing
from typing import Any, Generic, TypeVar, cast

from grpc import Channel
from packaging.version import Version
from packaging.version import parse as parse_version
from typing_extensions import Self

from ansys.api.acp.v0.base_pb2 import CollectionPath, DeleteRequest, GetRequest, ResourcePath

from .._server.acp_instance import ACPInstance, FileTransferHandler
from .._utils.path_to_str import path_to_str_checked
from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from .._utils.resource_paths import common_path
from .._utils.resource_paths import join as _rp_join
from .._utils.resource_paths import to_parts
from .._utils.typing_helper import PATH
from ._grpc_helpers.exceptions import wrap_grpc_errors
from ._grpc_helpers.linked_object_helpers import get_linked_paths, unlink_objects
from ._grpc_helpers.polymorphic_from_pb import (
    CreatableFromResourcePath,
    tree_object_from_resource_path,
)
from ._grpc_helpers.property_helper import (
    _exposed_grpc_property,
    _get_data_attribute,
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from ._grpc_helpers.protocols import (
    CreatableEditableAndReadableResourceStub,
    CreateRequest,
    Editable,
    EditableAndReadableResourceStub,
    GrpcObjectBase,
    ObjectInfo,
    Readable,
    ReadableResourceStub,
)
from ._object_cache import ObjectCacheMixin, constructor_with_cache


@mark_grpc_properties
class TreeObjectBase(ObjectCacheMixin, GrpcObjectBase):
    """Base class for ACP tree objects."""

    __slots__: Iterable[str] = ("_server_wrapper_store", "_pb_object")

    _COLLECTION_LABEL: str
    _OBJECT_INFO_TYPE: type[ObjectInfo]
    _SUPPORTED_SINCE: str

    _pb_object: ObjectInfo
    name: ReadOnlyProperty[str]

    def __init__(self: TreeObjectBase, name: str = "") -> None:
        self._server_wrapper_store: ServerWrapper | None = None
        self._pb_object: ObjectInfo = self._OBJECT_INFO_TYPE()
        # We don't want to invoke gRPC requests for setting the name
        # during object construction, so we set the name directly on
        # the protobuf object.
        self._pb_object.info.name = name

    @staticmethod
    def _cache_key_valid(key: Any) -> bool:
        if not isinstance(key, str):
            return False
        return bool(key)

    def __eq__(self: Self, other: Any) -> bool:
        if not isinstance(other, TreeObject):
            return False
        if not self._is_stored:
            # For unstored objects, fall back to identity comparison
            return self is other
        return self._resource_path.value == other._resource_path.value

    def __hash__(self) -> int:
        return id(self)

    @classmethod
    @constructor_with_cache(
        key_getter=lambda object_info, *args, **kwargs: object_info.info.resource_path.value,
        raise_on_invalid_key=False,
    )
    def _from_object_info(
        cls: type[Self], /, object_info: ObjectInfo, server_wrapper: ServerWrapper | None = None
    ) -> Self:
        instance = cls()
        instance._pb_object = object_info
        instance._server_wrapper_store = server_wrapper
        return instance

    @classmethod
    @constructor_with_cache(
        key_getter=lambda resource_path, *args, **kwargs: resource_path.value,
        raise_on_invalid_key=True,
    )
    def _from_resource_path(
        cls, /, resource_path: ResourcePath, server_wrapper: ServerWrapper
    ) -> Self:
        instance = cls()
        instance._pb_object.info.resource_path.CopyFrom(resource_path)
        instance._server_wrapper_store = server_wrapper
        return instance

    @property
    def _resource_path(self) -> ResourcePath:
        return self._pb_object.info.resource_path

    @property
    def _channel(self) -> Channel:
        return self._server_wrapper.channel

    @property
    def _server_wrapper(self) -> ServerWrapper:
        if not self._is_stored:
            raise RuntimeError("The server connection is uninitialized.")
        assert self._server_wrapper_store is not None
        return self._server_wrapper_store

    @property
    def _server_version(self) -> Version | None:
        if not self._is_stored:
            return None
        return self._server_wrapper.version

    @property
    def _is_stored(self) -> bool:
        return self._server_wrapper_store is not None

    @property
    def parent(self) -> CreatableFromResourcePath:
        """The parent of the object."""
        if not self._is_stored:
            raise RuntimeError("Cannot get the parent of an unstored object.")
        rp_parts = to_parts(self._resource_path.value)
        if len(rp_parts) < 3:
            raise RuntimeError("The object does not have a parent.")

        parent_path = _rp_join(*rp_parts[:-2])
        parent = tree_object_from_resource_path(
            ResourcePath(value=parent_path), server_wrapper=self._server_wrapper
        )
        if parent is None:
            raise RuntimeError("The parent object could not be found.")
        return parent

    def __repr__(self) -> str:
        return f"<{type(self).__name__} with name '{self.name}'>"

    name = grpc_data_property_read_only("info.name", doc="The name of the object.")


StubT = TypeVar("StubT")


@dataclass(frozen=True)
class ServerWrapper:
    """Wrapper for the connection to an ACP instance.

    This class contains the representation of the ACP instance needed by tree objects.
    Its purpose is to minimize the dependency of tree objects on the ACP class.
    """

    channel: Channel
    version: Version
    filetransfer_handler: FileTransferHandler

    @classmethod
    def from_acp_instance(cls, acp_instance: ACPInstance[Any]) -> ServerWrapper:
        """Convert an ACP instance into the wrapper needed by tree objects."""
        return cls(
            channel=acp_instance._channel,
            version=parse_version(acp_instance.server_version),
            filetransfer_handler=acp_instance._filetransfer_handler,
        )

    def auto_upload(self, local_path: PATH | None, allow_none: bool = False) -> str:
        """Handle auto-transfer of a file to the server."""
        if local_path is None:
            if allow_none:
                return ""
            raise TypeError("Expected a Path or str, not 'None'.")
        return path_to_str_checked(
            self.filetransfer_handler.upload_file_if_autotransfer(local_path)
        )

    @contextlib.contextmanager
    def auto_download(self, local_path: PATH) -> Iterator[str]:
        """Handle auto-transfer of a file from the server."""
        export_path = self.filetransfer_handler.to_export_path(local_path)
        yield path_to_str_checked(export_path)
        self.filetransfer_handler.download_file_if_autotransfer(export_path, local_path)


class StubStore(Generic[StubT]):
    """Stores a gRPC stub, and creates it on demand."""

    def __init__(self, create_stub_fun: Callable[[], StubT]) -> None:
        self._stub_store: StubT | None = None
        self._create_stub_fun = create_stub_fun

    def get(self, is_stored: bool) -> StubT:
        """Get the stored stub, or create it if it does not exist."""
        if not is_stored:
            raise RuntimeError("The server connection is uninitialized.")
        if self._stub_store is None:
            self._stub_store = self._create_stub_fun()
        return self._stub_store


class TreeObject(TreeObjectBase):
    """Base class for ACP objects which can be modified or deleted."""

    __slots__: Iterable[str] = ("_stub_store",)
    name: ReadWriteProperty[str, str] = grpc_data_property(
        "info.name", doc="The name of the object."
    )

    @abstractmethod
    def _create_stub(self) -> EditableAndReadableResourceStub: ...

    def __init__(self: TreeObject, name: str = "") -> None:
        super().__init__(name=name)
        self._stub_store = StubStore(self._create_stub)

    def delete(self) -> None:
        """Delete the object."""
        with wrap_grpc_errors():
            self._get_stub().Delete(
                DeleteRequest(
                    resource_path=self._pb_object.info.resource_path,
                    version=self._pb_object.info.version,
                )
            )

    def _get(self) -> None:
        with wrap_grpc_errors():
            self._pb_object = self._get_stub().Get(
                GetRequest(resource_path=self._pb_object.info.resource_path)
            )

    def _get_if_stored(self) -> None:
        if self._is_stored:
            self._get()

    def _put(self) -> None:
        with wrap_grpc_errors():
            self._pb_object = self._get_stub().Put(self._pb_object)

    def _put_if_stored(self) -> None:
        if self._is_stored:
            self._put()

    def _get_stub(self) -> EditableAndReadableResourceStub:
        return self._stub_store.get(self._is_stored)


@mark_grpc_properties
class ReadOnlyTreeObject(TreeObjectBase):
    """Base class for read-only ACP objects."""

    def __init__(self: ReadOnlyTreeObject) -> None:
        super().__init__()
        self._stub_store = StubStore(self._create_stub)

    @abstractmethod
    def _create_stub(self) -> ReadableResourceStub: ...

    def _get_stub(self) -> ReadableResourceStub:
        return self._stub_store.get(self._is_stored)

    def _get(self) -> None:
        with wrap_grpc_errors():
            self._pb_object = self._get_stub().Get(
                GetRequest(resource_path=self._pb_object.info.resource_path)
            )

    def _get_if_stored(self) -> None:
        if self._is_stored:
            self._get()


@mark_grpc_properties
class CreatableTreeObject(TreeObject):
    """Base class for ACP objects which can be created."""

    __slots__: Iterable[str] = tuple()
    _CREATE_REQUEST_TYPE: type[CreateRequest]

    def _get_stub(self) -> CreatableEditableAndReadableResourceStub:
        return cast(CreatableEditableAndReadableResourceStub, super()._get_stub())

    def clone(self: Self, *, unlink: bool = False) -> Self:
        """Create a new unstored object with the same properties.

        Parameters
        ----------
        unlink:
            If ``True``, remove all links to other objects. This can be
            used to store the object to another model, where the links
            would be invalid.
        """
        new_object_info = self._OBJECT_INFO_TYPE()
        new_object_info.properties.CopyFrom(self._pb_object.properties)
        if unlink:
            unlink_objects(new_object_info.properties)
            # Since there may be links in the unknown fields, we need to
            # discard them to avoid errors when storing the object.
            new_object_info.properties.DiscardUnknownFields()  # type: ignore
        new_object_info.info.name = self._pb_object.info.name
        return type(self)._from_object_info(object_info=new_object_info)

    def store(self: Self, parent: TreeObject) -> None:
        """Store the object on the server.

        Parameters
        ----------
        parent :
            Parent object to store the object under.
        """
        self._server_wrapper_store = parent._server_wrapper
        assert self._server_version is not None
        if self._server_version < parse_version(self._SUPPORTED_SINCE):
            raise RuntimeError(
                f"The '{type(self).__name__}' object is only supported since version "
                f"{self._SUPPORTED_SINCE} of the ACP gRPC server. The current server version is "
                f"{self._server_version}."
            )

        collection_path = CollectionPath(
            value=_rp_join(parent._resource_path.value, self._COLLECTION_LABEL)
        )

        # check that all linked objects are located in the same model
        path_values = [collection_path.value] + [
            path.value for path in get_linked_paths(self._pb_object.properties)
        ]
        # filter out empty paths
        path_values = [path for path in path_values if path]

        # Since the path starts with 'model/<model_uuid>', the objects belong to
        # the same model iff they share at least the first two parts.
        if len(to_parts(common_path(*path_values))) < 2:
            parent_model_uid = to_parts(parent._resource_path.value)[1]
            wrong_model_paths = []
            for linked_path in path_values:
                if to_parts(linked_path)[1] != parent_model_uid:
                    wrong_model_paths.append(linked_path)
            raise ValueError(
                "The object to store contains links to the following objects, which "
                + "are located on another Model: [\n    "
                + ",\n    ".join(repr(path) for path in wrong_model_paths)
                + "\n]"
            )

        request = self._CREATE_REQUEST_TYPE(
            collection_path=collection_path,
            name=self._pb_object.info.name,
            properties=self._pb_object.properties,
        )
        with wrap_grpc_errors():
            self._pb_object = self._get_stub().Create(request)
        resource_path_value = self._resource_path.value
        if not resource_path_value:
            raise ValueError("The resource path must not be empty.")
        self._OBJECT_CACHE[resource_path_value] = self


@mark_grpc_properties
class IdTreeObject(TreeObjectBase):
    """Implements the 'id' attribute for tree objects."""

    __slots__: Iterable[str] = tuple()

    id: ReadOnlyProperty[str] = grpc_data_property_read_only(
        "info.id", doc="Identifier of the object, used for example as key in maps."
    )

    def __repr__(self) -> str:
        return f"<{type(self).__name__} with id '{self.id}'>"


class TreeObjectAttributeReadOnly(GrpcObjectBase):
    """Read-only object backed by a sub-component of a parent object's protobuf object."""

    __slots__: Iterable[str] = ("_parent_object", "_attribute_path")

    def __init__(
        self,
        *,
        _parent_object: Readable | None = None,
        _attribute_path: str | None = None,
    ):
        if _parent_object is None != _attribute_path is None:
            raise TypeError(
                "Either both '_parent_object' and '_attribute_path' need to be 'None', or neither."
            )
        self._parent_object: Readable | None = _parent_object
        self._attribute_path = _attribute_path

    def _get(self) -> None:
        if self._parent_object is None:
            raise RuntimeError("The parent object is not set.")
        self._parent_object._get()

    def _get_if_stored(self) -> None:
        if self._is_stored:
            self._get()

    @property
    def _pb_object_impl(self) -> Any:
        assert self._parent_object is not None
        assert self._attribute_path is not None
        return _get_data_attribute(self._parent_object._pb_object, self._attribute_path)

    @property
    def _pb_object(self) -> Any:
        return self._pb_object_impl

    @property
    def _is_stored(self) -> bool:
        if self._parent_object is None:
            return False
        return self._parent_object._is_stored

    @property
    def _server_wrapper(self) -> ServerWrapper:
        if self._parent_object is None:
            raise RuntimeError("The parent object is not set.")
        return self._parent_object._server_wrapper


class PolymorphicMixin(TreeObjectAttributeReadOnly):
    """Mixin class for attributes which can have multiple types, through a 'oneof' definition."""

    @property
    def _pb_object_impl(self) -> Any:
        assert self._parent_object is not None
        assert self._attribute_path is not None
        *sub_path, prop_name = self._attribute_path.split(".")
        parent_attr = _get_data_attribute(self._parent_object._pb_object, ".".join(sub_path))
        return getattr(parent_attr, parent_attr.WhichOneof(prop_name))


class TreeObjectAttribute(TreeObjectAttributeReadOnly):
    """Editable object backed by a sub-component of a parent object's protobuf object."""

    __slots__: Iterable[str] = ("_parent_object", "_attribute_path", "_pb_object_store")

    @classmethod
    @abstractmethod
    def _create_default_pb_object(cls) -> Any: ...

    def __init__(
        self,
        *,
        _parent_object: Editable | None = None,
        _attribute_path: str | None = None,
        _pb_object: Any | None = None,
    ):
        if (_pb_object is not None) and (_parent_object is not None):
            raise TypeError("The '_pb_object' parameter is not allowed if '_parent_object' is set.")
        if _parent_object is None:
            if _pb_object is None:
                self._pb_object_store: Any = self._create_default_pb_object()
            else:
                self._pb_object_store = _pb_object
        else:
            self._pb_object_store = None
        self._parent_object: Editable | None
        super().__init__(_parent_object=_parent_object, _attribute_path=_attribute_path)

    @property
    def _pb_object(self) -> Any:
        if self._parent_object is None:
            return self._pb_object_store
        else:
            return self._pb_object_impl

    def _put(self) -> None:
        if self._parent_object is None:
            raise RuntimeError("The parent object is not set.")
        self._parent_object._put()

    def _put_if_stored(self) -> None:
        if self._is_stored:
            self._put()


class TreeObjectAttributeWithCache(ObjectCacheMixin, TreeObjectAttribute):
    """Tree object attribute with instance caching."""

    @classmethod
    @constructor_with_cache(
        key_getter=lambda parent_object, attribute_path: (
            parent_object,
            attribute_path,
        ),
        raise_on_invalid_key=True,
    )
    def _from_parent(cls: type[Self], /, parent_object: TreeObject, attribute_path: str) -> Self:
        return cls(_parent_object=parent_object, _attribute_path=attribute_path)

    @staticmethod
    def _cache_key_valid(key: tuple[Editable, str]) -> bool:
        parent_object, attribute_path = key
        return parent_object is not None and bool(attribute_path)


AttribT = TypeVar("AttribT", bound=TreeObjectAttributeWithCache)


def nested_grpc_object_property(
    pb_path: str,
    object_type: type[AttribT],
) -> ReadWriteProperty[AttribT, AttribT]:
    """Create a property for a nested object attribute.

    Creates a property for a nested object which is backed by the parent
    object's protobuf object. The property is read-write, and instances
    are cached.
    """

    def _getter(self: TreeObject) -> AttribT:
        return object_type._from_parent(parent_object=self, attribute_path=pb_path)

    def _setter(self: TreeObject, value: AttribT) -> None:
        if not isinstance(value, object_type):
            raise TypeError(
                f"Expected an object of type '{object_type.__name__}', got '{type(value).__name__}'."
            )
        _getter(self)._pb_object.CopyFrom(value._pb_object)
        self._put_if_stored()

    return _exposed_grpc_property(_getter).setter(_setter)


if typing.TYPE_CHECKING:  # pragma: no cover
    # Ensure that the ReadOnlyTreeObject satisfies the Gettable interface
    _x: Readable = typing.cast(ReadOnlyTreeObject, None)
    # Ensure that the TreeObject satisfies the Editable and Readable interfaces
    _y: Editable = typing.cast(TreeObject, None)
    _z: Readable = typing.cast(TreeObject, None)

    # Ensure the TreeObjectBase satisfies the CreatableFromResourcePath interface
    _a: CreatableFromResourcePath = typing.cast(TreeObjectBase, None)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

from ansys.api.acp.v0 import solid_model_export_pb2

from .._utils.typing_helper import PATH as _PATH
from ._grpc_helpers.exceptions import wrap_grpc_errors
from .base import CreatableTreeObject
from .enums import (
    SolidModelExportFormat,
    SolidModelSkinExportFormat,
    solid_model_export_format_to_pb,
    solid_model_skin_export_format_to_pb,
)

__all__ = ["SolidModelExportMixin"]


class SolidModelExportMixin(CreatableTreeObject):
    """Mixin class for adding export functionality to the solid model and imported solid model classes."""

    def export(self, path: _PATH, *, format: SolidModelExportFormat) -> None:
        """Export the solid model to a file.

        Parameters
        ----------
        path :
            Path to the file where the solid model is saved.
        format :
            Format of the exported file. Available formats are ``"ansys:h5"``
            and ``"ansys:cdb"``.

        """
        with self._server_wrapper.auto_download(path) as export_path:
            with wrap_grpc_errors():
                self._get_stub().ExportToFile(  # type: ignore
                    solid_model_export_pb2.ExportToFileRequest(
                        resource_path=self._resource_path,
                        path=export_path,
                        format=typing.cast(typing.Any, solid_model_export_format_to_pb(format)),
                    )
                )

    def export_skin(self, path: _PATH, *, format: SolidModelSkinExportFormat) -> None:
        """Export the skin of the solid model to a file.

        Parameters
        ----------
        path :
            Path to the file where the solid model skin is saved.
        format :
            Format of the exported file. Available formats are ``"ansys:cdb"``,
            ``"step"``, ``"iges"``, and ``"stl"``.

        """
        with self._server_wrapper.auto_download(path) as export_path:
            with wrap_grpc_errors():
                self._get_stub().ExportSkin(  # type: ignore
                    solid_model_export_pb2.ExportSkinRequest(
                        resource_path=self._resource_path,
                        path=export_path,
                        format=typing.cast(
                            typing.Any, solid_model_skin_export_format_to_pb(format)
                        ),
                    )
                )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable

from ansys.api.acp.v0 import rosette_pb2, rosette_pb2_grpc

from .._utils.array_conversions import to_1D_double_array, to_tuple_from_1D_array
from .._utils.property_protocols import ReadOnlyProperty
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .edge_set import EdgeSet
from .enums import RosetteType, rosette_type_from_pb, rosette_type_to_pb, status_type_from_pb
from .object_registry import register

__all__ = ["Rosette"]


@mark_grpc_properties
@register
class Rosette(CreatableTreeObject, IdTreeObject):
    """Instantiate a Rosette.

    Parameters
    ----------
    name :
        Name of the rosette.
    rosette_type :
        Type of the rosette.
    origin :
        Coordinates of the rosette origin.
    dir1 :
        Direction 1 (x-direction) vector of the Rosette.
    dir2 :
        Direction 2 (y-direction) vector of the Rosette.
    edge_set :
        Edge Set used for the Rosettes with type :attr:`RosetteType.EDGE_WISE`.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "rosettes"
    _OBJECT_INFO_TYPE = rosette_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = rosette_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "Rosette",
        rosette_type: RosetteType = RosetteType.PARALLEL,
        origin: tuple[float, float, float] = (0.0, 0.0, 0.0),
        dir1: tuple[float, float, float] = (1.0, 0.0, 0.0),
        dir2: tuple[float, float, float] = (0.0, 1.0, 0.0),
        edge_set: EdgeSet | None = None,
    ):
        super().__init__(name=name)

        self.rosette_type = rosette_type
        self.origin = origin
        self.dir1 = dir1
        self.dir2 = dir2
        self.edge_set = edge_set

    def _create_stub(self) -> rosette_pb2_grpc.ObjectServiceStub:
        return rosette_pb2_grpc.ObjectServiceStub(self._channel)

    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    origin = grpc_data_property(
        "properties.origin", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )
    dir1 = grpc_data_property(
        "properties.dir1", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )
    dir2 = grpc_data_property(
        "properties.dir2", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )

    rosette_type = grpc_data_property(
        "properties.rosette_type",
        from_protobuf=rosette_type_from_pb,
        to_protobuf=rosette_type_to_pb,
    )

    edge_set = grpc_link_property("properties.edge_set", allowed_types=EdgeSet)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable

from ansys.api.acp.v0 import fabric_pb2, fabric_pb2_grpc

from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import (
    CutOffMaterialHandling,
    DrapingMaterialModel,
    DropOffMaterialHandling,
    cut_off_material_type_from_pb,
    cut_off_material_type_to_pb,
    draping_material_type_from_pb,
    draping_material_type_to_pb,
    drop_off_material_type_from_pb,
    drop_off_material_type_to_pb,
    status_type_from_pb,
)
from .material import Material
from .object_registry import register

__all__ = ["Fabric"]


@mark_grpc_properties
@register
class Fabric(CreatableTreeObject, IdTreeObject):
    """Instantiate a Fabric.

    Parameters
    ----------
    name :
        Name of the fabric.
    material :
        Material of the fabric.
    thickness :
        Thickness of the fabric.
    area_price :
        Price per area of the fabric.
    ignore_for_postprocessing :
        Enable this option that the failure computation skips all plies made of this fabric.
    drop_off_material_handling :
        Defines the material of drop-off elements in the solid model extrusion.
    drop_off_material :
        Specify the material of drop-off elements in the solid model.
    cut_off_material_handling :
        Defines the material of cut-off elements in solid models if cut-off geometries are active.
    cut_off_material :
        Define the cut-off material if a ply with this material is shaped by a cut-off geometry.
    draping_material_model :
        Specifies the draping model of the fabric.
    draping_ud_coefficient :
        Set the draping coefficient of the uni-directional draping model. Must be in the range of 0 to 1.
    area_weight :
        Weight per area of the fabric. Read only.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "fabrics"
    _OBJECT_INFO_TYPE = fabric_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = fabric_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "Fabric",
        material: Material | None = None,
        thickness: float = 0.0,
        area_price: float = 0.0,
        ignore_for_postprocessing: bool = False,
        drop_off_material_handling: DropOffMaterialHandling = "global",
        drop_off_material: Material | None = None,
        cut_off_material_handling: CutOffMaterialHandling = "computed",
        cut_off_material: Material | None = None,
        draping_material_model: DrapingMaterialModel = "woven",
        draping_ud_coefficient: float = 0.0,
    ):
        super().__init__(name=name)

        self.material = material
        self.thickness = thickness
        self.area_price = area_price
        self.ignore_for_postprocessing = ignore_for_postprocessing
        self.drop_off_material_handling = DropOffMaterialHandling(drop_off_material_handling)
        self.drop_off_material = drop_off_material
        self.cut_off_material_handling = CutOffMaterialHandling(cut_off_material_handling)
        self.cut_off_material = cut_off_material
        self.draping_material_model = DrapingMaterialModel(draping_material_model)
        self.draping_ud_coefficient = draping_ud_coefficient

    def _create_stub(self) -> fabric_pb2_grpc.ObjectServiceStub:
        return fabric_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    area_weight: ReadOnlyProperty[float] = grpc_data_property_read_only("properties.area_weight")

    material = grpc_link_property("properties.material", allowed_types=Material)
    thickness: ReadWriteProperty[float, float] = grpc_data_property("properties.thickness")
    area_price: ReadWriteProperty[float, float] = grpc_data_property("properties.area_price")
    ignore_for_postprocessing: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.ignore_for_postprocessing"
    )

    drop_off_material_handling = grpc_data_property(
        "properties.drop_off_material_handling",
        from_protobuf=drop_off_material_type_from_pb,
        to_protobuf=drop_off_material_type_to_pb,
    )
    drop_off_material = grpc_link_property(
        "properties.drop_off_material",
        allowed_types=Material,
        readable_since="25.1",
        writable_since="25.1",
    )
    cut_off_material_handling = grpc_data_property(
        "properties.cut_off_material_handling",
        from_protobuf=cut_off_material_type_from_pb,
        to_protobuf=cut_off_material_type_to_pb,
    )
    cut_off_material = grpc_link_property(
        "properties.cut_off_material",
        allowed_types=Material,
        readable_since="25.1",
        writable_since="25.1",
    )

    draping_material_model = grpc_data_property(
        "properties.draping_material_model",
        from_protobuf=draping_material_type_from_pb,
        to_protobuf=draping_material_type_to_pb,
    )
    draping_ud_coefficient: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.draping_ud_coefficient"
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses
from typing import TYPE_CHECKING, cast

import numpy as np
import numpy.typing as npt
from typing_extensions import Self

if TYPE_CHECKING:  # pragma: no cover
    import pyvista

from ansys.api.acp.v0 import (
    base_pb2,
    cad_component_pb2_grpc,
    cad_geometry_pb2,
    cad_geometry_pb2_grpc,
)

from .._utils.array_conversions import to_numpy
from .._utils.path_to_str import path_to_str_checked
from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from .._utils.pyvista_import_check import requires_pyvista
from .._utils.typing_helper import PATH
from ._grpc_helpers.exceptions import wrap_grpc_errors
from ._grpc_helpers.mapping import get_read_only_collection_property
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .cad_component import CADComponent
from .enums import status_type_from_pb
from .object_registry import register


@dataclasses.dataclass
class TriangleMesh:
    """Represents a mesh consisting of triangle faces only."""

    node_coordinates: npt.NDArray[np.float64]
    element_nodes: npt.NDArray[np.int64]

    @classmethod
    def _from_pb(cls, response: cad_geometry_pb2.TriangleMeshData) -> Self:
        """Construct a triangle mesh object from a protobuf response."""

        return cls(
            to_numpy(response.node_coordinates),
            to_numpy(response.element_nodes),
        )

    @requires_pyvista
    def to_pyvista(
        self,
    ) -> pyvista.PolyData:
        """Convert the mesh data to a PyVista object."""
        import pyvista

        return pyvista.PolyData.from_regular_faces(  # type: ignore
            points=self.node_coordinates,
            faces=self.element_nodes,  # type: ignore[arg-type]
        )


@mark_grpc_properties
@register
class CADGeometry(CreatableTreeObject, IdTreeObject):
    """Instantiate a CAD Geometry.

    Parameters
    ----------
    name :
        Name of the CAD Geometry.
    external_path :
        Path to the CAD file.
    scale_factor :
        Scale factor applied to the CAD geometry.
    use_default_precision :
        Whether to use the default precision value.
    precision :
        Precision of geometrical operations such as intersection points, thickness sampling, etc.
    use_default_offset :
        Whether to use the default offset value.
    offset :
        Offset value used to analyze the surface's coverage regarding the mesh.
    """

    __slots__: Iterable[str] = tuple()
    _COLLECTION_LABEL = "cad_geometries"
    _OBJECT_INFO_TYPE = cad_geometry_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = cad_geometry_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "CADGeometry",
        external_path: str = "",
        scale_factor: float = 1.0,
        use_default_precision: bool = True,
        precision: float = 1e-3,
        use_default_offset: bool = True,
        offset: float = 0.0,
    ):
        super().__init__(
            name=name,
        )
        self.external_path = external_path
        self.scale_factor = scale_factor
        self.use_default_precision = use_default_precision
        self.precision = precision
        self.use_default_offset = use_default_offset
        self.offset = offset

    def _create_stub(self) -> cad_geometry_pb2_grpc.ObjectServiceStub:
        return cad_geometry_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")

    external_path: ReadWriteProperty[PATH, PATH] = grpc_data_property(
        "properties.external_path", to_protobuf=path_to_str_checked
    )
    scale_factor: ReadWriteProperty[float, float] = grpc_data_property("properties.scale_factor")
    use_default_precision: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_default_precision"
    )
    precision: ReadWriteProperty[float, float] = grpc_data_property("properties.precision")
    use_default_offset: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_default_offset"
    )
    offset: ReadWriteProperty[float, float] = grpc_data_property("properties.offset")

    @property
    def visualization_mesh(self) -> TriangleMesh:
        """Surface mesh for visualization.

        The CAD Geometry's surface represented as a triangle mesh.
        """
        if not self._is_stored:
            raise RuntimeError("Cannot get mesh data from an unstored object")
        stub = cast(cad_geometry_pb2_grpc.ObjectServiceStub, self._get_stub())
        with wrap_grpc_errors():
            response = stub.GetMesh(
                request=base_pb2.GetRequest(
                    resource_path=self._resource_path,
                ),
            )
        return TriangleMesh._from_pb(response)

    root_shapes = get_read_only_collection_property(
        object_class=CADComponent,
        stub_class=cad_component_pb2_grpc.ObjectServiceStub,
        requires_uptodate=True,
    )

    def refresh(self, path: PATH) -> None:
        """Reload the geometry from its external source.

        Parameters
        ----------
        path :
            Path of the new input file.
        """
        self.external_path = self._server_wrapper.auto_upload(path)
        stub = cast(cad_geometry_pb2_grpc.ObjectServiceStub, self._get_stub())
        with wrap_grpc_errors():
            stub.Refresh(
                request=cad_geometry_pb2.RefreshRequest(
                    resource_path=self._resource_path,
                ),
            )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helper functions for exchanging data between PyACP and PyMechanical."""

import pathlib
import shutil
import textwrap
import typing

if typing.TYPE_CHECKING:
    import ansys.mechanical.core as pymechanical

from ._utils.typing_helper import PATH

__all__ = [
    "export_mesh_for_acp",
    "import_acp_composite_definitions",
    "import_acp_mesh_from_cdb",
]


def export_mesh_for_acp(*, mechanical: "pymechanical.Mechanical", path: PATH) -> None:
    """Export the mesh from PyMechanical for use in PyACP.

    Parameters
    ----------
    mechanical :
        The PyMechanical instance. This must be a remote instance.
    path :
        The path to save the mesh to. The extension must be '.h5'.
    """
    path = pathlib.Path(path)

    if path.suffix != ".h5":
        raise ValueError(f"The output path extension must be '.h5', not '{path.suffix}'.")
    output_path_str = str(path)
    mechanical.run_python_script(
        textwrap.dedent(
            f"""\
            geometry_type = Ansys.Mechanical.DataModel.Enums.GeometryType.Sheet
            unit = Ansys.Mechanical.DataModel.Enums.WBUnitSystemType.ConsistentMKS
            dsid = 0

            Model.InternalObject.WriteHDF5TransferFile(geometry_type, {output_path_str!r}, unit, dsid)
            """
        )
    )


def import_acp_mesh_from_cdb(*, mechanical: "pymechanical.Mechanical", cdb_path: PATH) -> None:
    """Import an ACP CDB mesh into Mechanical.

    Import a mesh exported from ACP in CDB format into Mechanical. This function
    does not import the ACP layup definition, use :func:`import_acp_composite_definitions`
    for this purpose.

    .. warning::

        The named selections exported from ACP are only partially imported.

    Parameters
    ----------
    mechanical :
        The PyMechanical instance. This must be a remote instance.
    cdb_path :
        The path of the CDB file to import. The extension must be '.cdb'.
    """
    cdb_path = pathlib.Path(cdb_path)

    if cdb_path.suffix != ".cdb":
        raise ValueError(f"The CDB file extension must be '.cdb', not '{cdb_path.suffix}'.")
    cdb_path_str = str(cdb_path)

    mechanical.run_python_script(
        textwrap.dedent(
            f"""\
            model_import = Model.AddGeometryImportGroup().AddModelImport()
            model_import.ModelImportSourceFilePath = {cdb_path_str!r}
            model_import.ProcessValidBlockedCDBFile = False
            model_import.ProcessModelData = False
            model_import.Import()
            """
        )
    )


def import_acp_composite_definitions(*, mechanical: "pymechanical.Mechanical", path: PATH) -> None:
    """Import ACP composite definitions HDF5 into Mechanical.

    Imports the composite layup defined in ACP into Mechanical, as Imported Plies.

    This function does not import the solid mesh, use :func:`import_acp_mesh_from_cdb`
    for this purpose.

    Parameters
    ----------
    mechanical :
        The PyMechanical instance. This must be a remote instance.
    path :
        The path of the file to import. The extension must be '.h5'.
    """
    path = pathlib.Path(path)

    if path.suffix != ".h5":
        raise ValueError(
            f"The composite definitions file extension must be '.h5', not '{path.suffix}'."
        )

    user_files_dir = pathlib.Path(mechanical.project_directory) / "UserFiles"
    user_files_dir.mkdir(exist_ok=True)
    target_path = user_files_dir / path.name
    try:
        shutil.copyfile(path, target_path)
    except shutil.SameFileError:
        pass

    target_path_str = f"Setup::{str(target_path.resolve())}"

    mechanical.run_python_script(
        textwrap.dedent(
            f"""\
            import clr
            clr.AddReference("Ansys.Common.Interop.{mechanical.version}")
            composite_definition_paths_coll = Ansys.Common.Interop.AnsCoreObjects.AnsBSTRColl()
            composite_definition_paths_coll.Add({target_path_str!r})
            mapping_paths_coll = Ansys.Common.Interop.AnsCoreObjects.AnsVARIANTColl()
            mapping_paths_coll.Add(None)
            external_model = Model.InternalObject.AddExternalEnhancedModel(
                Ansys.Common.Interop.DSObjectTypes.DSExternalEnhancedModelType.kEXTERNAL_ENHANCEDMODEL_LAYEREDSECTION
            )
            external_model.Import(composite_definition_paths_coll, mapping_paths_coll)
            external_model.Update()
            """
        )
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Iterable

from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.message import Message

from ansys.api.acp.v0.base_pb2 import ResourcePath

__all__ = ("unlink_objects", "get_linked_paths")


def unlink_objects(pb_object: Message) -> None:
    """Remove all ResourcePaths and CollectionPaths from a protobuf object."""
    for parent_message, field_descriptor, _ in _linked_path_fields(pb_object):
        parent_message.ClearField(field_descriptor.name)


def get_linked_paths(pb_object: Message) -> Iterable[ResourcePath]:
    """Get all resource paths present in a protobuf object."""
    for _, field_descriptor, field_value in _linked_path_fields(pb_object):
        if field_descriptor.label == field_descriptor.LABEL_REPEATED:
            yield from field_value  # type: ignore
        else:
            yield field_value  # type: ignore


def _linked_path_fields(
    pb_object: Message,
) -> Iterable[tuple[Message, FieldDescriptor, Message]]:
    """Get the field field information for resource paths in the message.

    Get tuples (parent_message, field_descriptor, field_value) describing
    all resource paths present in the protobuf object. Note that the fields
    can also be repeated (containing multiple resource paths).
    """
    for field_descriptor, field_value in pb_object.ListFields():
        if getattr(field_descriptor.message_type, "name", None) == "ResourcePath":
            yield (pb_object, field_descriptor, field_value)
        elif field_descriptor.type == field_descriptor.TYPE_MESSAGE:
            if field_descriptor.label == field_descriptor.LABEL_REPEATED:
                for sub_obj in field_value:
                    yield from _linked_path_fields(sub_obj)
            else:
                yield from _linked_path_fields(field_value)
